"""Containers for various types of waveforms

Currently, the most interesting object in this submodule is [`WaveformModes`](../sxs.waveforms.waveform_modes/).

"""

from .waveform_mixin import WaveformMixin
from .waveform_modes import WaveformModes
# from .waveform_grid import WaveformGrid
# from .waveform_signal import WaveformSignal

from . import nrar, rotating_paired_xor_multishuffle_bzip2, memory, transformations #, corotating_paired_xor

formats = {
    None: nrar,
    "": nrar,
    "nrar": nrar,
    "rotating_paired_xor_multishuffle_bzip2": rotating_paired_xor_multishuffle_bzip2,
    # "corotating_paired_xor": corotating_paired_xor,
}



import os
import functools
import pytest
import numpy as np
import quaternionic

ell_max_default = 36

try:
    import spinsfast
    requires_spinsfast = lambda f: f
except:
    requires_spinsfast = pytest.mark.skip(reason="spinsfast is missing")

try:
    import sympy
    requires_sympy = lambda f: f
except:
    requires_sympy = pytest.mark.skip(reason="sympy is missing")



def pytest_addoption(parser):
    parser.addoption("--ell_max", action="store", type=int, default=ell_max_default,
                     help="Maximum ell value to test")
    parser.addoption("--ell_max_slow", action="store", type=int, default=ell_max_default // 2,
                     help="Maximum ell value to test with slow tests")
    parser.addoption("--run_slow_tests", action="store_true", default=False,
                     help="Run all tests, including slow ones")


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: marks tests as slow")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--run_slow_tests"):
        return
    skip_slow = pytest.mark.skip(reason="need --run_slow_tests option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


def pytest_runtest_setup(item):
    if 'slow' in item.keywords and not item.config.getoption("--run_slow_tests"):
        pytest.skip("Need `--run_slow_tests` command-line argument to run")


@pytest.fixture
def ell_max(request):
    return request.config.getoption("--ell_max")


@pytest.fixture
def ell_max_slow(request):
    return request.config.getoption("--ell_max_slow")


@pytest.fixture
def special_angles():
    return np.arange(-1 * np.pi, 1 * np.pi + 0.1, np.pi / 4.)


@pytest.fixture
def on_windows():
    from sys import platform
    return 'win' in platform.lower() and not 'darwin' in platform.lower()


@pytest.fixture
def eps():
    return np.finfo(float).eps


def quaternion_sampler():
    Qs_array = quaternionic.array([
        [np.nan, 0., 0., 0.],
        [np.inf, 0., 0., 0.],
        [-np.inf, 0., 0., 0.],
        [0., 0., 0., 0.],
        [1., 0., 0., 0.],
        [0., 1., 0., 0.],
        [0., 0., 1., 0.],
        [0., 0., 0., 1.],
        [1.1, 2.2, 3.3, 4.4],
        [-1.1, -2.2, -3.3, -4.4],
        [1.1, -2.2, -3.3, -4.4],
        [
            0.18257418583505537115232326093360,
            0.36514837167011074230464652186720,
            0.54772255750516611345696978280080,
            0.73029674334022148460929304373440
        ],
        [1.7959088706354, 0.515190292664085, 0.772785438996128, 1.03038058532817],
        [2.81211398529184, -0.392521193481878, -0.588781790222817, -0.785042386963756],
    ])
    names = type("QNames", (object,), dict())()
    names.q_nan1 = 0
    names.q_inf1 = 1
    names.q_minf1 = 2
    names.q_0 = 3
    names.q_1 = 4
    names.x = 5
    names.y = 6
    names.z = 7
    names.Q = 8
    names.Qneg = 9
    names.Qbar = 10
    names.Qnormalized = 11
    names.Qlog = 12
    names.Qexp = 13
    return Qs_array, names


@pytest.fixture
def Qs():
    return quaternion_sampler()[0]


@pytest.fixture
def Q_names():
    return quaternion_sampler()[1]


@pytest.fixture
def Q_conditions():
    Qs_array, names = quaternion_sampler()
    conditions = type("QConditions", (object,), dict())()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        conditions.zero = np.arange(len(Qs_array))[Qs_array == Qs_array[names.q_0]]
        conditions.nonzero = np.arange(len(Qs_array))[np.nonzero(Qs_array)]
        conditions.nan = np.arange(len(Qs_array))[np.isnan(Qs_array)]
        conditions.nonnan = np.arange(len(Qs_array))[~np.isnan(Qs_array)]
        conditions.nonnannonzero = np.arange(len(Qs_array))[~np.isnan(Qs_array) & (Qs_array != Qs_array[names.q_0])]
        conditions.inf = np.arange(len(Qs_array))[np.isinf(Qs_array)]
        conditions.noninf = np.arange(len(Qs_array))[~np.isinf(Qs_array)]
        conditions.noninfnonzero = np.arange(len(Qs_array))[~np.isinf(Qs_array) & (Qs_array != Qs_array[names.q_0])]
        conditions.finite = np.arange(len(Qs_array))[np.isfinite(Qs_array)]
        conditions.nonfinite = np.arange(len(Qs_array))[~np.isfinite(Qs_array)]
        conditions.finitenonzero = np.arange(len(Qs_array))[np.isfinite(Qs_array) & (Qs_array != Qs_array[names.q_0])]
    return conditions


@pytest.fixture
def Rs():
    np.random.seed(1842)
    ones = [0, -1., 1.]
    rs = [[w, x, y, z] for w in ones for x in ones for y in ones for z in ones][1:]
    rs = rs + [r for r in [quaternionic.array(np.random.uniform(-1, 1, size=4)) for _ in range(20)]]
    return quaternionic.array(rs).normalized


# catalog = sxs.load("catalog")
# com_files = [
#     f for f in catalog.files.values()
#     if "rhOverM_Asymptotic_GeometricUnits_CoM.h5" in f['filename']
#     and "SXS:BBH:1111v" not in f['truepath']
# ]
# min(com_files, key=lambda f: f['filesize'])
shortest_h_com_file = "SXS:BBH:0156v1/Lev5/rhOverM_Asymptotic_GeometricUnits_CoM.h5"
shortest_horizons = "SXS:BBH:0156v1/Lev5/Horizons.h5"
shortest_metadata = "SXS:BBH:0156v5/Lev5/metadata.json"
shortest_metadata_txt = "SXS:BBH:0156v5/Lev5/metadata.txt"


@functools.lru_cache()
def get_h():
    import contextlib
    import sxs
    with contextlib.redirect_stdout(None):
        h = sxs.load(shortest_h_com_file, extrapolation_order=3)
    return h


@pytest.fixture
def h():
    return get_h().copy()



"""Functions to enable BMS transformations of waveforms

BMS transformations include the usual Poincaré group — time and space
translations, rotations, and boosts — as well as "supertranslations", which are
a more general form of translations.  Essentially, a supertranslation is a
direction-dependent time translation.

See https://arxiv.org/abs/1509.00862 for a review of BMS transformations and
their computation.

"""


def uprime_generator(u, β):
    """Return u' such that each time step is the smallest in the input time series

    Parameters
    ----------
    u : array_like
        Time steps in the original (stationary) frame
    β : float
        Magnitude of boost velocity as fraction of the speed of light

    Returns
    -------
    uprime : array_like
        Time steps in the boosted frame

    Notes
    -----
    The u' time steps in the boosted frame incorporate data from a range of
    different time slices in the stationary frame.  Because the size of time steps
    can vary as a function of time to represent dynamical data adaptively, the
    timestep sizes of those different slices will differ.  Here, we construct new
    times for the u' series so that the u' steps are no larger than the smallest
    time step on any of those slices from the input data.

    We simplify this somewhat by assuming that the time step sizes in the input `u`
    are fairly smoothly varying, so that the appropriate time step can be
    determined just by looking at the earliest and latest time slices.  This is
    only an approximation, but should be suitable for our data.

    """
    import numpy as np
    from scipy.interpolate import CubicSpline
    # uprime = u / (γ * (1 - v⃗ · n̂))
    # uprime_min = max(min(u) / (γ * (1 - v⃗ · n̂)))
    # uprime_max = min(max(u) / (γ * (1 - v⃗ · n̂)))

    γ = 1 / np.sqrt(1 - β**2)
    uprime_min = max(min(u) / (γ * (1 - β)), min(u) / (γ * (1 + β)))
    uprime_max = min(max(u) / (γ * (1 - β)), max(u) / (γ * (1 + β)))
    if uprime_max < uprime_min:
        raise ValueError(
            f"\n\tThere are no complete slices in the u' coordinate system for u ∈ [{min(u)}, {max(u)}] and β = {β}."
            f"\n\tYou may wish to decrease β or move the origin of the time coordinate closer to (u[0] + u[-1]) / 2."
        )
    uprime = [uprime_min,]
    δuprime_plus = CubicSpline(u[1:], np.diff(u / (γ * (1 + β))), extrapolate=True)
    δuprime_minus = CubicSpline(u[1:], np.diff(u / (γ * (1 - β))), extrapolate=True)
    while uprime[-1] < uprime_max:
        δuprime = min(
            δuprime_plus(uprime[-1] * γ * (1 + β)),
            δuprime_minus(uprime[-1] * γ * (1 - β))
        )
        uprime.append(min(uprime_max, uprime[-1] + δuprime))
    uprime = np.array(uprime)
    return uprime


def Bprime(v⃗, n̂prime):
    """Rotor of aberration spin-weighted fields under boosts

    Implements Equation (2) of arxiv.org/abs/1509.00862

    Parameters
    ----------
    v⃗ : (3,) array_like
        Three-vector representing the velocity of the boosted frame relative to the
        inertial frame, in units where the speed of light is 1
    n̂prime : (..., 3) array_like
        Three-vectors representing the directions in the boosted frame

    Returns
    -------
    Bprm : (..., 4) quaternionic.array
        Quaternions that rotate from the boosted frame to the stationary frame.  In
        addition to rotating n̂prime onto the corresponding n̂ directions, this will
        also rotate tangent vectors appropriately, to account for spin
        transformations.  The shape of this array is n̂prime.shape[:-1]+(4,).

    """
    import numpy as np
    import quaternionic
    ϵ = 1e-15
    β = np.linalg.norm(v⃗)
    if β < ϵ:
        return quaternionic.one
    v̂ = v⃗ / β
    φ = np.arctanh(β)
    Θprime = np.arccos(np.tensordot(v̂, n̂prime, axes=[-1, -1]))
    Θ = 2 * np.arctan(np.exp(-φ) * np.tan(Θprime/2))
    nv = np.cross(n̂prime, v̂)
    nvnorm = np.linalg.norm(nv, axis=-1)
    nv /= nvnorm[..., np.newaxis]
    return np.exp(((Θprime - Θ) / 2) * quaternionic.array.from_vector_part(nv))


def boost(w, v⃗, ell_max):
    """Find modes of waveform boosted by velocity v⃗

    Implements Equation (21) of arxiv.org/abs/1509.00862

    Parameters
    ----------
    w : WaveformModes
        Modes of waveform measured in original frame
    v⃗ : array_like
        Three-vector representing the velocity of the boosted frame relative to the
        inertial frame, in units where the speed of light is 1
    ell_max : int
        Maximum value of `ell` to use while computing the transformation, and to
        provide in the returned object

    Returns
    -------
    wprime : WaveformModes
        Modes of waveform measured in boosted frame or of modes from boosted source
        measured in original frame.  This should have the same properties as the
        input `w`, except with (1) different time data [see Notes, below], (2) a
        minimum `ell` value of 0 even for spin weight other than 0, and (3) a
        maximum `ell` value of `ell_max`.

    Notes
    -----
    Due to the nature of the transformation, some of the information in the input
    waveform must be discarded, because it corresponds to slices of the output
    waveform that are not completely represented in the input.  Thus, the times of
    the output waveform will not just be the Lorentz-transformed times of the input
    waveform.

    Depending on the magnitude β=|v⃗|, a very large value of `ell_max` may be
    needed.  The dominant factor is the translation that builds up over time:
    `β*T`, where `T` is the largest time found in the waveform.  For example, if
    β*T ≈ 1000M, we might need `ell_max=64` to maintain a comparable accuracy as in
    the input data.

    Because of the `β*T` effects, it is usually best to set t=0 at the merger time
    — best approximated as `self.max_norm_time()`.  The largest translation is then
    found early in the waveform, when the waveform is changing slowly.

    """
    import numpy as np
    import quaternionic
    import spherical
    import spinsfast

    if w.data_type.lower() not in ['h', 'psi4']:
        raise NotImplementedError(f"Input waveform `w` has type {w.data_type}, which is not yet implemented")

    ϵ = 1e-15
    β = np.linalg.norm(v⃗)

    if β < ϵ:
        return w.copy()

    γ = 1 / np.sqrt(1 - β**2)
    φ = np.arctanh(β)
    v̂ = v⃗ / β

    nθprime = nϕprime = 2 * ell_max + 1
    θprimeϕprime = spherical.theta_phi(nθprime, nϕprime)
    Rprime = quaternionic.array.from_spherical_coordinates(θprimeφprime)
    n̂prime = (Rprime * quaternionic.z * Rprime.inverse).vector

    R = Bprime(v⃗, n̂prime) * Rprime
    n̂ = (R * quaternionic.z * R.inverse).vector
    doppler_factor = γ * (1 - np.tensordot(v⃗, n̂, axes=[-1, -1]))

    uprime = uprime_generator(w.t, β)
    time_axis, modes_axis = 0, 1
    Hprime = np.zeros((uprime.size, spherical.Ysize(0, ell_max)), dtype=complex)

    # Step through the waveform in segments, so that each segment is small enough to fit
    # comfortably into memory, but large enough to minimize re-computation of SWSHs
    i_step_size = 5_000
    i_padding = 20
    i_outer_1 = 0
    i_inner_1 = 0
    i_inner_2 = min(i_inner_1 + i_step_size, uprime.size)
    i_outer_2 = min(i_inner_2 + i_padding, uprime.size)
    Hprime_grid = np.zeros((i_step_size, nθprime, nϕprime), dtype=complex)
    while True:
        uprime_outer = uprime[i_outer_1:i_outer_2]
        uprime_inner = uprime[i_inner_1:i_inner_2]

        # print(f"Working on ({uprime_inner[[0, -1]]}) of ({uprime[[0, -1]]})")

        # Within each segment, this is the core computation, evaluating the transformed
        # field on a grid, and then converting back to mode weights
        for j in range(Rprime.shape[0]):
            for k in range(Rprime.shape[1]):
                Rjk = R[j, k]
                doppler_factor_jk = doppler_factor[j, k]
                u_outer_1, u_outer_2 = uprime_outer[[0, -1]] * doppler_factor_jk
                i1, i2 = w.index_closest_to(u_outer_1), w.index_closest_to(u_outer_2)
                Hprime_grid[:i_inner_2-i_inner_1, j, k] = (
                    doppler_factor_jk * w[i1:i2].evaluate(Rjk).interpolate(uprime_inner * doppler_factor_jk)
                )
        Hprime[i_inner_1:i_inner_2] = spinsfast.map2salm(Hprime_grid[:i_inner_2-i_inner_1], w.spin_weight, ell_max)

        # Move to the next segment
        if i_inner_2 == uprime.size:
            break
        i_inner_1 = i_inner_2
        i_outer_1 = max(0, i_inner_1 - i_padding)
        i_inner_2 = min(i_inner_2 + i_step_size, uprime.size)
        i_outer_2 = min(i_inner_2 + i_padding, uprime.size)

    return type(w)(
        Hprime, time=uprime, time_axis=time_axis, modes_axis=modes_axis,
        ell_min=0, ell_max=ell_max, data_type=w.data_type, spin_weight=w.spin_weight
    )



"""Search NASA's ADS system

Documentation can be found here: <http://adsabs.github.io/help/>

"""


api_url = 'https://api.adsabs.harvard.edu/v1/'


class Login(object):

    def __init__(self, api_token=None, api_token_path=None,
                 total_retry_count=50, backoff_factor=0.1, backoff_max=20.0, session=None):
        """Initialize a Login object for interacting with NASA ADS

        This object encapsulates the credentials needed to interact with the NASA/ADS API, and
        exposes a Session object that can be used to make requests which automatically include the
        credentials.

        These actions require a NASA/ADS API access token.  These are obtained from the website:
            https://ui.adsabs.harvard.edu/user/settings/token
        The access token may either be passed as a string to this function (though that means it
        will probably be found in some script file somewhere, which is probably not a good idea for
        security), or can be read from a file.  By default the file from which the token is read is
        '~/.credentials/ads/api_token'.  Thus, it is probably easiest to simply place your access
        tokens in those files, so that no arguments need to be passed to this function.  As a basic
        security measure, please ensure that those files are not readable by anyone but the user.


        Parameters
        ==========
        api_token: string or None [default: None]
            If present, this is used as the NASA/ADS API access token.

        api_token_path: string or None [default: None]
            If `api_token` is not given, this file is read and the first line is used as the access
            token.  If this argument is None, it defaults to '~/.credentials/ads/api_token'.

        total_retry_count: int [default: 50]
            Total number of times to retry requests that fail for retry-able reasons.

        backoff_factor: float [default: 0.1]
            A delay factor to apply between requests after the second try (most errors are resolved
            immediately by a second try without a delay).  After a certain number of total retries,
            the request Session will sleep for:

                {backoff factor} * (2 ^ ({number of total retries} - 1))

            seconds before trying again. For example, if the `backoff_factor` is 0.1, then the
            session will sleep for [0.0s, 0.2s, 0.4s, ...] between retries.  It will never be longer
            than `backoff_max`.

        backoff_max: float [default: 20.0]
            Longest time (in seconds) to wait between retries.

        session: requests.Session or None [default: None]
            This is the object that handles all of the requests made to the API.  If `None`, a
            Session is created for you, and sensible default headers (including the access token)
            are created.  If you need to adjust some of the Session parameters like proxies or SSL
            verification, you can simply create your own and pass it in here.  Note that any `auth`
            property on the passed object will be replaced by one that adds this to the header of
            each request to the NASA/ADS domain:
                {"Authorization": "Bearer <YourAccessTokenHere>"}

        """
        import os
        import requests
        from requests.adapters import HTTPAdapter
        from requests.packages.urllib3.util.retry import Retry

        self.base_url = api_url

        # The Session object will handle all requests we make.
        self.session = session or requests.Session()

        # If the input session object succeeds at a `get`, we can skip a lot of the following
        r = self.session.get(self.base_url+'search/query', params={'q': 'star'})
        if r.status_code != 200:  # That's okay, we mostly expected this

            # Set the NASA/ADS API access token
            if api_token is not None:
                self.api_token = api_token
            else:
                if api_token_path is None:
                    api_token_path = '~/.credentials/ads/api_token'
                path = os.path.expanduser(api_token_path)
                try:
                    with open(path, 'r') as f:
                        self.api_token = f.readline().strip()
                except IOError:
                    print('Unable to find the NASA/ADS access token needed to query the system.')
                    print('Failed to open file "{0}" for reading.'.format(path))
                    raise
                if not self.api_token:
                    print('The file "{0}" did not contain any text on the first line.'.format(path))
                    print('This is should be a NASA/ADS access token, which is need to query the system.')
                    raise ValueError('Query requires a NASA/ADS access token')

            # Ensure that this session sends the Authorization header with every request to the base_url
            class NASAADSAuth(requests.auth.AuthBase):
                def __init__(self, base_url, api_token):
                    self.base_url = base_url
                    self.api_token = api_token
                    super(NASAADSAuth, self).__init__()
                def __call__(self, r):
                    if r.url.startswith(self.base_url):
                        r.headers.update({"Authorization": "Bearer {0}".format(self.api_token)})
                    return r
            self.session.auth = NASAADSAuth(self.base_url, self.api_token)

        # Note that some requests require different choices for 'Accept' and 'Content-Type'; these
        # are altered in the corresponding methods below.
        default_headers= {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        self.session.headers.update(default_headers)

        ## Retry automatically on certain types of errors
        Retry.BACKOFF_MAX = backoff_max  # Must be set on the class, not the instance
        retry = Retry(
            total=total_retry_count,
            backoff_factor=backoff_factor,
            status_forcelist=[500, 502, 503, 504,],
        )
        adapter = HTTPAdapter(max_retries=retry)
        self.session.mount(self.base_url, adapter)

        # Test to see if we can use the given access token
        r = self.session.get(self.base_url+'search/query', params={'q': 'star'})
        if r.status_code != 200:
            if r.status_code == 401:
                print('The given NASA/ADS access token was not accepted by {0}.  Please ensure that it is still valid.'.format(self.base_url))
            else:
                print('An unknown error occurred when trying to access {0}.'.format(self.base_url))
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error

    @property
    def export_formats(self):
        """List of formats in which ADS can export records"""
        return ['bibtex', 'bibtexabs', 'ads', 'endnote', 'procite', 'ris', 'refworks', 'rss', 'medlars',
                'dcxml', 'refxml', 'refabsxml', 'aastex', 'icarus', 'mnras', 'soph', 'votable', 'custom']

    def export(self, format='bibtex', ):
        pass

    def query(self, query_dict):
        query_url = api_url + 'search/query'
        r = self.session.get(query_url, params={'q': query_dict})
        if r.status_code != 200:
            print('An error occurred when trying to access <{0}>.'.format(query_url), file=sys.stderr)
            try:
                print(r.json(), file=sys.stderr)
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error
        return r.json()


def ads_bearer_token(error_if_not_found=True):
    import os.environ
    if 'ADS_TOKEN' not in os.environ:
        message = 'Environment variable "ADS_TOKEN" not found'
        if error_if_not_found:
            raise EnvironmentError(message)
        else:
            from warnings import warn
            warn(message)
            return ''
    return os.environ('ADS_TOKEN')


def authorization_header(error_if_not_found=True):
    token = ads_bearer_token(error_if_not_found)
    return 'Authorization: Bearer:{0}'.format(token)



#curl -H 'Authorization: Bearer:Co3woTdgh00kgY46RrP51sVHZBFEiml6LeQq2xNK' 'http://api.adsabs.harvard.edu/v1/search/query?q=doi:10.1111/j.1365-2966.2010.18127.x&fl=pub+volume+issue+year+page'



"""Handlers for waveforms in the old NRAR data format"""

import sys
import warnings
import ast
import numpy as np
from . import WaveformModes
from ..metadata import Metadata
from ..utilities.monotonicity import index_is_monotonic


FrameType = [UnknownFrameType, Inertial, Coprecessing, Coorbital, Corotating] = range(5)
FrameNames = ["UnknownFrameType", "Inertial", "Coprecessing", "Coorbital", "Corotating"]

DataType = [UnknownDataType, psi0, psi1, psi2, psi3, psi4, sigma, h, hdot, news, psin] = range(11)
DataNames = ["UnknownDataType", "Psi0", "Psi1", "Psi2", "Psi3", "Psi4", "sigma", "h", "hdot", "news", "psin"]
SpinWeights = [sys.maxsize, 2, 1, 0, -1, -2, 2, -2, -2, -2, sys.maxsize]
ConformalWeights = [sys.maxsize, 2, 1, 0, -1, -2, 1, 0, -1, -1, -3]
RScaling = [sys.maxsize, 5, 4, 3, 2, 1, 2, 1, 1, 1, 0]
MScaling = [sys.maxsize, 2, 2, 2, 2, 2, 0, 0, 1, 1, 2]
DataNamesLaTeX = [
    r"\mathrm{unknown data type}",
    r"\psi_0",
    r"\psi_1",
    r"\psi_2",
    r"\psi_3",
    r"\psi_4",
    r"\sigma",
    r"h",
    r"\dot{h}",
    r"\mathrm{n}",
    r"\psi_n",
]


def translate_date_type_to_spin_weight(d):
    return [sys.maxsize, 2, 1, 0, -1, -2, 2, -2, -2, -2, sys.maxsize][d]

def translate_date_type_to_sxs_string(d):
    return ["unknown", "psi0", "psi1", "psi2", "psi3", "psi4", "sigma", "h", "hdot", "news", "psin"][d]

def translate_data_types_GWFrames_to_waveforms(d):
    if d < 8:
        return {0: UnknownDataType, 1: h, 2: hdot, 3: psi4, 4: psi3, 5: psi2, 6: psi1, 7: psi0}[d]
    else:
        return DataType[d - 8]


def translate_data_types_waveforms_to_GWFrames(d):
    if d in [UnknownDataType, h, hdot, psi4, psi3, psi2, psi1, psi0]:
        return {UnknownDataType: 0, h: 1, hdot: 2, psi4: 3, psi3: 4, psi2: 5, psi1: 6, psi0: 7}[d]
    else:
        return d + 8


def frame_type_string(w):
    return FrameNames[w._metadata.get("frame_type", UnknownFrameType)]


def data_type_string(w):
    return DataNames[w._metadata.get("data_type", UnknownDataType)]


def data_type_latex(w):
    return DataNamesLaTeX[w._metadata.get("data_type", UnknownDataType)]


def descriptor_string(w):
    """Create a simple string describing the content of the waveform

    This string will be suitable for file names.  For example, 'rMpsi4' or
    'rhOverM'.  It uses the waveform's knowledge of itself, so if this is
    incorrect, the result will be incorrect.

    """
    dataType = w._metadata.get("data_type", UnknownDataType)
    r_is_scaled_out = w._metadata.get("r_is_scaled_out", 1)
    m_is_scaled_out = w._metadata.get("m_is_scaled_out", 1)
    data_type_string = DataNames[dataType]
    if dataType == UnknownDataType:
        return data_type_string
    descriptor = ""
    if r_is_scaled_out:
        if RScaling[dataType] == 1:
            descriptor = "r"
        elif RScaling[dataType] > 1:
            descriptor = "r" + str(RScaling[dataType])
    if m_is_scaled_out:
        Mexponent = MScaling[dataType] - (RScaling[dataType] if r_is_scaled_out else 0)
        if Mexponent < -1:
            descriptor = descriptor + data_type_string + "OverM" + str(-Mexponent)
        elif Mexponent == -1:
            descriptor = descriptor + data_type_string + "OverM"
        elif Mexponent == 0:
            descriptor = descriptor + data_type_string
        elif Mexponent == 1:
            descriptor = descriptor + "M" + data_type_string
        elif Mexponent > 1:
            descriptor = descriptor + "M" + str(Mexponent) + data_type_string
    else:
        descriptor = descriptor + data_type_string
    return descriptor


def load(file, **kwargs):
    """Read data from an H5 file in NRAR format

    Note that either `h5_group` or `extrapolation_order` must be given.

    Parameters
    ----------
    file : {str, pathlib.Path}
        Path to H5 file containing the data, optionally including the path within
        the file itself to the directory containing the data.  For example, the
        standard SXS data with N=2 might be obtained with the file name
        `'rhOverM_Asymptotic_GeometricUnits.h5/Extrapolated_N2.dir'`.

    Keyword parameters
    ------------------
    h5_group : str, optional
        HDF5 group in which to find the data within the H5 file
    extrapolation_order : {None, str, int, Ellipsis}, optional
        Extrapolation order to load from the file.  This can be a string naming the
        HDF5 group to find the data in — like "OutermostExtraction.dir" or
        "Extrapolated_N2.dir" — or it can be an integer that is translated into
        such a string, where -1 is translated into "OutermostExtraction.dir".  The
        default value of `None` will return a dict-like object mapping names like
        the above to their corresponding waveforms, but will also issue a warning
        that it is doing so.  The final option of `Ellipsis` (which can also be
        input as `...`) will also return this dict-like object, but will not issue
        a warning.
    frame_type : int, optional
    data_type : int, optional
    r_is_scaled_out : bool, optional
        True if the radius is scaled out, meaning that the asymptotic form of this
        data could be finite and nonzero.
    m_is_scaled_out : bool, optional
        True if the mass is scaled out of the data, so that the data does not
        change when the total mass changes (presumably for vacuum systems only).

    """
    import pathlib
    import re
    import h5py
    import quaternionic
    import spherical
    from ..utilities import KeyPassingDict

    # This unfortunate concoction is needed to determine the (ell,m) values of the various mode data sets
    pattern_Ylm = re.compile(r"""Y_l(?P<L>[0-9]+)_m(?P<M>[-+0-9]+)\.dat""")

    # Initialize an empty object to be filled with goodies
    w_attributes = {}

    # Get an h5py handle to the desired part of the h5 file
    file_str = str(file)
    split = file_str.rsplit(".h5", 1)  # Raises ValueError if ".h5" is not in the string
    if len(split) != 2:
        split = file_str.rsplit(".hdf5", 1)
        file_ending = ".hdf5"
    else:
        file_ending = ".h5"
    if len(split) != 2:
        raise ValueError(f"Could not find a valid HDF5 filename ending in '{file_str}'")
    file_str, root_group = split
    file_str = file_str + file_ending
    if not root_group:
        root_group = kwargs.pop("h5_group", "")
    if not root_group:
        extrapolation_order = kwargs.pop("extrapolation_order", None)
        if extrapolation_order is None:
            warning = "\nCould not find root group as `h5_group` or as `extrapolation_order`; returning all groups"
            warnings.warn(warning)
        elif extrapolation_order is Ellipsis:
            pass
        elif isinstance(extrapolation_order, str):
            root_group = extrapolation_order
        elif extrapolation_order == -1:
            root_group = "OutermostExtraction.dir"
        else:
            root_group = f"Extrapolated_N{extrapolation_order}.dir"

    file_path = pathlib.Path(file_str).expanduser().resolve()
    file_name = file_path.name
    file_dir = file_path.parent

    with h5py.File(file_path, "r") as f_h5:
        if root_group:
            if root_group not in f_h5:
                raise ValueError(f"Input root group '{root_group}' was not found in '{file_path}'")
            f = f_h5[root_group]
        else:
            return KeyPassingDict(**{
                new_h5_group: load(file, **dict(h5_group=new_h5_group, **kwargs))
                for new_h5_group in f_h5 if "VersionHist.ver" not in new_h5_group
            }, **{
                new_h5_group: f_h5[new_h5_group][:]
                for new_h5_group in f_h5 if "VersionHist.ver" in new_h5_group
            })

        # If it exists, add the metadata file to `w` as an object.  So, for example, the initial spin on
        # object 1 can be accessed as `w.metadata.initial_spin1`.  See the documentation of
        # `sxs.metadata.Metadata` for more details.  And in IPython, tab completion works on the
        # `w.metadata` object.
        try:
            w_attributes["metadata"] = Metadata.from_file(
                file_dir / "metadata", ignore_invalid_lines=True, cache_json=False
            )
        except:
            pass  # Probably couldn't find the metadata.json/metadata.txt file

        try:  # Print a more explanatory error message if this fails
            # Extract the old History to search for COM-correction parameters
            if "History.txt" in f:
                try:
                    old_history = f["History.txt"][()].decode()
                except AttributeError:
                    old_history = f["History.txt"][()]
            else:
                old_history = ""
            w_attributes["history"] = old_history

            # Get the frame data, converting to quaternion objects
            if "Frame" in f:
                w_attributes["frame"] = quaternionic.array(f["Frame"])

            # Get the descriptive items
            if "FrameType" in f.attrs:
                w_attributes["frame_type"] = int(f.attrs["FrameType"])
            elif "frame_type" in kwargs:
                w_attributes["frame_type"] = int(kwargs.pop("frame_type"))
            else:
                warning = (
                    f"\n`frameType` was not found in '{file_str}' or the keyword arguments.\n"
                    + "Using default value `{}`.  You may want to set it manually.\n\n".format(FrameNames[1])
                )
                warnings.warn(warning)
                w_attributes["frame_type"] = 1

            if "DataType" in f.attrs:
                w_attributes["data_type"] = translate_data_types_GWFrames_to_waveforms(int(f.attrs["DataType"]))
            elif "data_type" in kwargs:
                w_attributes["data_type"] = int(kwargs.pop("data_type"))
            elif any(file_name.startswith(s) for s in {"rhOverM_", "rh_"}):
                w_attributes["data_type"] = h
            elif any(file_name.startswith(s) for s in {"rMPsi4_", "rPsi4_"}):
                w_attributes["data_type"] = psi4
            else:
                found = False
                for type_int, type_name in zip(reversed(DataType), reversed(DataNames)):
                    if type_name.lower() in file_name.lower():
                        found = True
                        w_attributes["data_type"] = type_int
                        break
                if not found:
                    warning = (
                        f"\n`dataType` was not found in '{file_str}' or the keyword arguments.\n"
                        + f"Using default value `{type_name}`.  You may want to set it manually."
                    )
                    warnings.warn(warning)
                    w_attributes["data_type"] = 0
            w_attributes["spin_weight"] = translate_date_type_to_spin_weight(w_attributes["data_type"])
            w_attributes["data_type"] = translate_date_type_to_sxs_string(w_attributes["data_type"])

            if "RIsScaledOut" in f.attrs:
                w_attributes["r_is_scaled_out"] = bool(f.attrs["RIsScaledOut"])
            elif "r_is_scaled_out" in kwargs:
                w_attributes["r_is_scaled_out"] = bool(kwargs.pop("r_is_scaled_out"))
            elif any(file_name.startswith(s) for s in {"rhOverM_", "rh_", "rMPsi4_", "rPsi4_"}):
                w_attributes["r_is_scaled_out"] = True
            else:
                warning = (
                    f"\n`r_is_scaled_out` was not found in '{file_str}' or the keyword arguments.\n"
                    + "Using default value `True`.  You may want to set it manually.\n\n"
                )
                warnings.warn(warning)
                w_attributes["r_is_scaled_out"] = True

            if "MIsScaledOut" in f.attrs:
                w_attributes["m_is_scaled_out"] = bool(f.attrs["MIsScaledOut"])
            elif "m_is_scaled_out" in kwargs:
                w_attributes["m_is_scaled_out"] = bool(kwargs.pop("m_is_scaled_out"))
            elif any(file_name.startswith(s) for s in {"rhOverM_", "rMPsi4_"}):
                w_attributes["m_is_scaled_out"] = True
            else:
                warning = (
                    f"\n`m_is_scaled_out` was not found in '{file_str}' or the keyword arguments.\n"
                    + "Using default value `True`.  You may want to set it manually.\n\n"
                )
                warnings.warn(warning)
                w_attributes["m_is_scaled_out"] = True

            # Get the names of all the data sets in the h5 file, and check for matches
            YLMdata = [data_set for data_set in list(f) for m in [pattern_Ylm.search(data_set)] if m]
            if len(YLMdata) == 0:
                raise ValueError(
                    f"Couldn't understand data set names in '{file_str}'.\n"
                    + "Maybe you need to add the directory within the h5 file.\n"
                    + f"E.g.: '{file_str}/Extrapolated_N2.dir'."
                )

            # Sort the data set names by increasing ell, then increasing m
            YLMdata = sorted(
                YLMdata,
                key=lambda data_set: [
                    int(pattern_Ylm.search(data_set).group("L")),
                    int(pattern_Ylm.search(data_set).group("M")),
                ],
            )
            LM = np.array(
                sorted(
                    [
                        [int(m.group("L")), int(m.group("M"))]
                        for data_set in YLMdata
                        for m in [pattern_Ylm.search(data_set)]
                        if m
                    ]
                )
            )
            ell_min, ell_max = min(LM[:, 0]), max(LM[:, 0])
            if not np.array_equal(LM, spherical.Yrange(ell_min, ell_max)):
                raise ValueError(f"Input [ell,m] modes are not complete.  Found modes:\n{LM}\n")
            n_modes = len(LM)

            # Get the time data (assuming all are equal)
            T = f[YLMdata[0]][:, 0]
            monotonic = index_is_monotonic(T)
            w_attributes["t"] = T[monotonic]
            n_times = len(w_attributes["t"])

            # Loop through, setting data in each mode
            w_attributes["data"] = np.empty((n_times, n_modes), dtype=complex)
            for m, DataSet in enumerate(YLMdata):
                if f[DataSet].shape[0] != n_times:
                    raise ValueError(
                        f"The number of time steps in this dataset should be {n_times}; "
                        + "it is {} in '{}'.".format(f[DataSet].shape[0], DataSet)
                    )
                w_attributes["data"][:, m] = f[DataSet][:, 1:3].view(dtype=np.complex)[monotonic, 0]

            # Now that the data is set, we can set these
            w_attributes["ells"] = ell_min, ell_max

            # If possible, retrieve the CoM-correction parameters
            try:
                if hasattr(f, "attrs") and "space_translation" in f.attrs:
                    w_attributes["space_translation"] = np.array(list(f.attrs["space_translation"]))
                elif old_history:
                    pattern = r"'{}': array\((.*?)\)".format("space_translation")
                    matches = re.search(pattern, old_history)
                    if matches:
                        w_attributes["space_translation"] = np.array(ast.literal_eval(matches.group(1)))
            except:
                pass
            try:
                if hasattr(f, "attrs") and "boost_velocity" in f.attrs:
                    w_attributes["boost_velocity"] = np.array(list(f.attrs["boost_velocity"]))
                elif old_history:
                    pattern = r"'{}': array\((.*?)\)".format("boost_velocity")
                    matches = re.search(pattern, old_history)
                    if matches:
                        w_attributes["boost_velocity"] = np.array(ast.literal_eval(matches.group(1)))
            except:
                pass

            # If possible, retrieve the CoM-correction parameters
            try:
                if "VersionHist.ver" in f_h5:
                    w_attributes["version_hist"] = [
                        [git_hash.decode("ascii"), message.decode("ascii")]
                        for git_hash, message in f_h5["VersionHist.ver"][()].tolist()
                    ]
            except:
                pass

        except KeyError as e:
            raise ValueError("\nThis H5 file appears to have not stored all the required information.\n\n") from e

    if kwargs:
        import pprint
        warnings.warn("\nUnused kwargs passed to this function:\n{}".format(pprint.pformat(kwargs, width=1)))

    data = w_attributes.pop("data")
    time = w_attributes.pop("t")
    ell_min, ell_max = w_attributes.pop("ells")
    metadata = w_attributes.pop("metadata", {})
    w = WaveformModes(
        data,
        time=time,
        time_axis=0,
        modes_axis=1,
        ell_min=ell_min,
        ell_max=ell_max,
        **w_attributes
    )
    if metadata:
        w.metadata = metadata

    return w


def save(w, file_name, file_write_mode="w", attributes={}, use_NRAR_format=True):
    """Output to an HDF5 file

    Note that the file_name is prepended with some descriptive information
    involving the data type and the frame type, such as 'rhOverM_Corotating_' or
    'rMpsi4_Aligned_'.

    """

    import os.path
    import h5py
    import warnings

    group = None
    if ".h5" in file_name and not file_name.endswith(".h5"):
        file_name, group = file_name.rsplit(".h5", 1)
        file_name += ".h5"

    # Add descriptive prefix to file_name
    base_name = descriptor_string(w) + "_" + os.path.basename(file_name)
    if not os.path.dirname(file_name):
        file_name = base_name
    else:
        file_name = os.path.join(os.path.dirname(file_name), base_name)

    # Open the file for output
    with h5py.File(file_name, file_write_mode) as f:
        # If we are writing to a group within the file, create it
        if group:
            g = f.create_group(group)
        else:
            g = f

        # Now write all the data to various groups in the file
        g.attrs["OutputFormatVersion"] = "scri.SpEC"
        g.attrs["FrameType"] = w._metadata["frame_type"]
        g.attrs["DataType"] = translate_data_types_waveforms_to_GWFrames(w._metadata["data_type"])
        g.attrs["RIsScaledOut"] = int(w._metadata.get("r_is_scaled_out", 1))
        g.attrs["MIsScaledOut"] = int(w._metadata.get("m_is_scaled_out", 1))
        if "version_hist" in w._metadata:
            try:
                version_hist = [[e.encode("ascii", "ignore") for e in hm] for hm in w._metadata["version_hist"]]
            except AttributeError:
                version_hist = w._metadata["version_hist"]
            g.create_dataset("VersionHist.ver", data=version_hist)
        for attr in attributes:
            try:
                g.attrs[attr] = attributes[attr]
            except:
                warning = f"scri.SpEC.write_to_h5 unable to output attribute {attr}={attributes[attr]}"
                warnings.warn(warning)
        if use_NRAR_format:
            for i_m in range(w.n_modes):
                ell, m = w.LM[i_m]
                Data_m = g.create_dataset(
                    f"Y_l{ell}_m{m}.dat",
                    data=[[t, d.real, d.imag] for t, d in zip(w.time, w.data[:, i_m])],
                    compression="gzip",
                    shuffle=True,
                )
                Data_m.attrs["ell"] = ell
                Data_m.attrs["m"] = m
        else:
            g.create_dataset("Time", data=w.time.tolist(), compression="gzip", shuffle=True)
            frame = w._metadata.get("frame", [])
            if len(frame) > 0:
                g.create_dataset("Frame", data=[[r.w, r.x, r.y, r.z] for r in frame])
            else:
                g.create_dataset("Frame", shape=())
            Data = g.create_group("Data")
            for i_m in range(w.n_modes):
                ell, m = w.LM[i_m]
                Data_m = Data.create_dataset(
                    "l{}_m{:+}".format(int(ell), int(m)), data=w.data[:, i_m], compression="gzip", shuffle=True
                )
                Data_m.attrs["ell"] = ell
                Data_m.attrs["m"] = m



def limit_author_list(all_authors, desired_authors=[], author_list_length_limit=10):
    if len(all_authors) <= author_list_length_limit:
        return all_authors
    author_list = [a for a in all_authors if a in desired_authors]
    if not author_list:
        author_list = all_authors[:author_list_length_limit]
    if len(author_list) < len(all_authors):
        author_list += ['et~al.']
    return author_list



import numpy as np
import sxs
import pytest


def test_uprime_generator():
    β = 0.1
    for t1, t2 in [[-30, -10], [-10, 10], [10, 30]]:
        u = np.linspace(t1, t2, num=1000)
        uprime = sxs.waveforms.transformations.uprime_generator(u, β)
        assert uprime[0] > u[0] and uprime[-1] < u[-1]
    for t1, t2 in [[-20, 0]]:
        u = np.linspace(t1, t2, num=1000)
        uprime = sxs.waveforms.transformations.uprime_generator(u, β)
        assert uprime[0] > u[0] and uprime[-1] == u[-1]
    for t1, t2 in [[0, 20]]:
        u = np.linspace(t1, t2, num=1000)
        uprime = sxs.waveforms.transformations.uprime_generator(u, β)
        assert uprime[0] == u[0] and uprime[-1] < u[-1]

    β = 0.9
    for t1, t2 in [[-10, 10]]:
        u = np.linspace(t1, t2, num=1000)
        uprime = sxs.waveforms.transformations.uprime_generator(u, β)
        assert uprime[0] >= u[0] and uprime[-1] <= u[-1]
    for t1, t2 in [[-20, 0]]:
        u = np.linspace(t1, t2, num=1000)
        uprime = sxs.waveforms.transformations.uprime_generator(u, β)
        assert uprime[0] > u[0] and uprime[-1] == u[-1]
    for t1, t2 in [[0, 20]]:
        u = np.linspace(t1, t2, num=1000)
        uprime = sxs.waveforms.transformations.uprime_generator(u, β)
        assert uprime[0] == u[0] and uprime[-1] < u[-1]
    for t1, t2 in [[-10, -9], [9, 10]]:
        u = np.linspace(t1, t2, num=1000)
        with pytest.raises(ValueError):
            uprime = sxs.waveforms.transformations.uprime_generator(u, β)
